// D import file generated from 'gl3n/linalg.d'
module gl3n.linalg;
private 
{
	import std.math : isNaN, isInfinity;
	import std.conv : to;
	import std.traits : isIntegral, isFloatingPoint, isStaticArray, isDynamicArray, isImplicitlyConvertible, isArray;
	import std.string : format, rightJustify;
	import std.array : join;
	import std.algorithm : max, min, reduce;
	import gl3n.math : clamp, PI, sqrt, sin, cos, acos, tan, asin, atan2, almost_equal;
	import gl3n.util : is_vector, is_matrix, is_quaternion, TupleRange;
}
version (NoReciprocalMul)
{
	private enum rmul = false;
}
else
{
	private enum rmul = true;
}
struct Vector(type, int dimension_)
{
	static assert(dimension > 0, "0 dimensional vectors don't exist.");
	alias vt = type;
	static const int dimension = dimension_;
	vt[dimension] vector;
	auto const @property value_ptr()
	{
		return vector.ptr;
	}
	@property string as_string()
	{
		return format("%s", vector);
	}
	alias toString = as_string;
	nothrow pure @safe 
	{
		private inout @property ref inout(vt) get_(char coord)()
		{
			return vector[coord_to_index!coord];
		}
		alias x = get_!'x';
		alias u = x;
		alias s = x;
		alias r = x;
		static if (dimension >= 2)
		{
			alias y = get_!'y';
			alias v = y;
			alias t = y;
			alias g = y;
		}
		static if (dimension >= 3)
		{
			alias z = get_!'z';
			alias b = z;
			alias p = z;
		}
		static if (dimension >= 4)
		{
			alias w = get_!'w';
			alias a = w;
			alias q = w;
		}
		static if (dimension == 2)
		{
			enum Vector e1 = Vector(1.to!vt, 0.to!vt);
			enum Vector e2 = Vector(0.to!vt, 1.to!vt);
		}
		else
		{
			static if (dimension == 3)
			{
				enum Vector e1 = Vector(1.to!vt, 0.to!vt, 0.to!vt);
				enum Vector e2 = Vector(0.to!vt, 1.to!vt, 0.to!vt);
				enum Vector e3 = Vector(0.to!vt, 0.to!vt, 1.to!vt);
			}
			else
			{
				static if (dimension == 4)
				{
					enum Vector e1 = Vector(1.to!vt, 0.to!vt, 0.to!vt, 0.to!vt);
					enum Vector e2 = Vector(0.to!vt, 1.to!vt, 0.to!vt, 0.to!vt);
					enum Vector e3 = Vector(0.to!vt, 0.to!vt, 1.to!vt, 0.to!vt);
					enum Vector e4 = Vector(0.to!vt, 0.to!vt, 0.to!vt, 1.to!vt);
				}
			}
		}
		static void isCompatibleVectorImpl(int d)(Vector!(vt, d) vec) if (d <= dimension)
		{
		}
		enum isCompatibleVector(T) = is(typeof(isCompatibleVectorImpl(T.init)));
		static void isCompatibleMatrixImpl(int r, int c)(Matrix!(vt, r, c) m)
		{
		}
		enum isCompatibleMatrix(T) = is(typeof(isCompatibleMatrixImpl(T.init)));
		private void construct(int i, T, Tail...)(T head, Tail tail)
		{
			static if (i >= dimension)
			{
				static assert(false, "Too many arguments passed to constructor");
			}
			else
			{
				static if (is(T : vt))
				{
					vector[i] = head;
					construct!(i + 1)(tail);
				}
				else
				{
					static if (isDynamicArray!T)
					{
						static assert(Tail.length == 0 && i == 0, "dynamic array can not be passed together with other arguments");
						vector[] = head[];
					}
					else
					{
						static if (isStaticArray!T)
						{
							vector[i..i + T.length] = head[];
							construct!(i + T.length)(tail);
						}
						else
						{
							static if (isCompatibleVector!T)
							{
								vector[i..i + T.dimension] = head.vector[];
								construct!(i + T.dimension)(tail);
							}
							else
							{
								static assert(false, "Vector constructor argument must be of type " ~ vt.stringof ~ " or Vector, not " ~ T.stringof);
							}
						}
					}
				}
			}
		}
		private void construct(int i)()
		{
			static assert(i == dimension, "Not enough arguments passed to constructor");
		}
		this(Args...)(Args args)
		{
			construct!0(args);
		}
		this(T)(T vec) if (is_vector!T && is(T.vt : vt) && T.dimension >= dimension)
		{
			foreach (i; TupleRange!(0, dimension))
			{
				vector[i] = vec.vector[i];
			}
		}
		this()(vt value)
		{
			clear(value);
		}
		const @property bool isFinite()
		{
			static if (isIntegral!type)
			{
				return true;
			}
			else
			{
				foreach (v; vector)
				{
					if (isNaN(v) || isInfinity(v))
					{
						return false;
					}
				}
				return true;
			}
		}
		deprecated("Use isFinite instead of ok") alias ok = isFinite;
		void clear(vt value)
		{
			foreach (i; TupleRange!(0, dimension))
			{
				vector[i] = value;
			}
		}
		template coord_to_index(char c)
		{
			static if (c == 'x' || c == 'r' || c == 'u' || c == 's')
			{
				enum coord_to_index = 0;
			}
			else
			{
				static if (c == 'y' || c == 'g' || c == 'v' || c == 't')
				{
					enum coord_to_index = 1;
				}
				else
				{
					static if (c == 'z' || c == 'b' || c == 'p')
					{
						static assert(dimension >= 3, "the " ~ c ~ " property is only available on vectors with a third dimension.");
						enum coord_to_index = 2;
					}
					else
					{
						static if (c == 'w' || c == 'a' || c == 'q')
						{
							static assert(dimension >= 4, "the " ~ c ~ " property is only available on vectors with a fourth dimension.");
							enum coord_to_index = 3;
						}
						else
						{
							static assert(false, "accepted coordinates are x, s, r, u, y, g, t, v, z, p, b, w, q and a not " ~ c ~ ".");
						}
					}
				}
			}
		}
		static if (dimension == 2)
		{
			void set(vt x, vt y)
			{
				vector[0] = x;
				vector[1] = y;
			}
		}
		static if (dimension == 3)
		{
			void set(vt x, vt y, vt z)
			{
				vector[0] = x;
				vector[1] = y;
				vector[2] = z;
			}
		}
		static if (dimension == 4)
		{
			void set(vt x, vt y, vt z, vt w)
			{
				vector[0] = x;
				vector[1] = y;
				vector[2] = z;
				vector[3] = w;
			}
		}
		void update(Vector!(vt, dimension) other)
		{
			vector = other.vector;
		}
		private const void dispatchImpl(int i, string s, int size)(ref vt[size] result)
		{
			static if (s.length > 0)
			{
				result[i] = vector[coord_to_index!(s[0])];
				dispatchImpl!(i + 1, s[1 .. $])(result);
			}

		}
		const @property Vector!(vt, s.length) opDispatch(string s)()
		{
			vt[s.length] ret;
			dispatchImpl!(0, s)(ret);
			Vector!(vt, s.length) ret_vec;
			ret_vec.vector = ret;
			return ret_vec;
		}
		const @property real magnitude_squared()
		{
			real temp = 0;
			foreach (index; TupleRange!(0, dimension))
			{
				temp += vector[index] ^^ 2;
			}
			return temp;
		}
		const @property real magnitude()
		{
			return sqrt(magnitude_squared);
		}
		alias length_squared = magnitude_squared;
		alias length = magnitude;
		void normalize()
		{
			real len = length;
			if (len != 0)
			{
				foreach (index; TupleRange!(0, dimension))
				{
					vector[index] = cast(type)(vector[index] / len);
				}
			}
		}
		const @property Vector normalized()
		{
			Vector ret;
			ret.update(this);
			ret.normalize();
			return ret;
		}
		const Vector opUnary(string op : "-")()
		{
			Vector ret;
			foreach (index; TupleRange!(0, dimension))
			{
				ret.vector[index] = -vector[index];
			}
			return ret;
		}
		const Vector opBinary(string op : "*")(vt r)
		{
			Vector ret;
			foreach (index; TupleRange!(0, dimension))
			{
				ret.vector[index] = vector[index] * r;
			}
			return ret;
		}
		const Vector opBinary(string op : "/")(vt r)
		{
			Vector ret;
			foreach (index; TupleRange!(0, dimension))
			{
				ret.vector[index] = vector[index] / r;
			}
			return ret;
		}
		const Vector opBinary(string op)(Vector r) if (op == "+" || op == "-")
		{
			Vector ret;
			foreach (index; TupleRange!(0, dimension))
			{
				ret.vector[index] = mixin("vector[index]" ~ op ~ "r.vector[index]");
			}
			return ret;
		}
		const vt opBinary(string op : "*")(Vector r)
		{
			return dot(this, r);
		}
		const Vector!(vt, T.cols) opBinary(string op : "*", T)(T inp) if (isCompatibleMatrix!T && T.rows == dimension)
		{
			Vector!(vt, T.cols) ret;
			ret.clear(0);
			foreach (c; TupleRange!(0, T.cols))
			{
				foreach (r; TupleRange!(0, T.rows))
				{
					ret.vector[c] += vector[r] * inp.matrix[r][c];
				}
			}
			return ret;
		}
		auto const opBinaryRight(string op, T)(T inp) if (!is_vector!T && !is_matrix!T && !is_quaternion!T)
		{
			return this.opBinary!op(inp);
		}
		void opOpAssign(string op : "*")(vt r)
		{
			foreach (index; TupleRange!(0, dimension))
			{
				vector[index] *= r;
			}
		}
		void opOpAssign(string op : "/")(vt r)
		{
			foreach (index; TupleRange!(0, dimension))
			{
				vector[index] /= r;
			}
		}
		void opOpAssign(string op)(Vector r) if (op == "+" || op == "-")
		{
			foreach (index; TupleRange!(0, dimension))
			{
				mixin("vector[index]" ~ op ~ "= r.vector[index];");
			}
		}
		const int opCmp(ref const Vector vec)
		{
			foreach (i, a; vector)
			{
				if (a < vec.vector[i])
				{
					return -1;
				}
				else
					if (a > vec.vector[i])
					{
						return 1;
					}
			}
			return 0;
		}
		const bool opEquals(T)(const T vec) if (!isArray!T && T.dimension == dimension)
		{
			return vector == vec.vector;
		}
		const bool opEquals(T)(const(T)[] array) if (!isArray!T && !is_vector!T)
		{
			if (array.length != dimension)
			{
				return false;
			}
			foreach (index; TupleRange!(0, dimension))
			{
				if (vector[index] != array[index])
				{
					return false;
				}
			}
			return true;
		}
		const bool opCast(T : bool)()
		{
			return isFinite;
		}
	}
}
pure nothrow @safe T.vt dot(T)(const T veca, const T vecb) if (is_vector!T)
{
	T.vt temp = 0;
	foreach (index; TupleRange!(0, T.dimension))
	{
		temp += veca.vector[index] * vecb.vector[index];
	}
	return temp;
}
pure nothrow @safe T cross(T)(const T veca, const T vecb) if (is_vector!T && T.dimension == 3)
{
	return T(veca.y * vecb.z - vecb.y * veca.z, veca.z * vecb.x - vecb.z * veca.x, veca.x * vecb.y - vecb.x * veca.y);
}
pure nothrow @safe T.vt distance(T)(const T veca, const T vecb) if (is_vector!T)
{
	return (veca - vecb).length;
}
pure nothrow @safe T reflect(T)(const T vec, const T norm) if (is_vector!T)
{
	return 2 * (vec * norm) * norm - vec;
}
alias vec2 = Vector!(float, 2);
alias vec3 = Vector!(float, 3);
alias vec4 = Vector!(float, 4);
alias vec2d = Vector!(double, 2);
alias vec3d = Vector!(double, 3);
alias vec4d = Vector!(double, 4);
alias vec2i = Vector!(int, 2);
alias vec3i = Vector!(int, 3);
alias vec4i = Vector!(int, 4);
struct Matrix(type, int rows_, int cols_) if (rows_ > 0 && cols_ > 0)
{
	alias mt = type;
	static const int rows = rows_;
	static const int cols = cols_;
	mt[cols][rows] matrix;
	alias matrix this;
auto const @property value_ptr()
	{
		return matrix[0].ptr;
	}
	@property string as_string()
	{
		return format("%s", matrix);
	}
	alias toString = as_string;
	@property string as_pretty_string()
	{
		string fmtr = "%s";
		size_t rjust = max(format(fmtr, reduce!max(matrix[])).length, format(fmtr, reduce!min(matrix[])).length) - 1;
		string[] outer_parts;
		foreach (mt[] row; matrix)
		{
			string[] inner_parts;
			foreach (mt col; row)
			{
				inner_parts ~= rightJustify(format(fmtr, col), rjust);
			}
			outer_parts ~= " [" ~ join(inner_parts, ", ") ~ "]";
		}
		return "[" ~ join(outer_parts, "\x0a")[1..$] ~ "]";
	}
	alias toPrettyString = as_pretty_string;
	nothrow pure @safe 
	{
		static void isCompatibleMatrixImpl(int r, int c)(Matrix!(mt, r, c) m)
		{
		}
		enum isCompatibleMatrix(T) = is(typeof(isCompatibleMatrixImpl(T.init)));
		static void isCompatibleVectorImpl(int d)(Vector!(mt, d) vec)
		{
		}
		enum isCompatibleVector(T) = is(typeof(isCompatibleVectorImpl(T.init)));
		private void construct(int i, T, Tail...)(T head, Tail tail)
		{
			static if (i >= rows * cols)
			{
				static assert(false, "Too many arguments passed to constructor");
			}
			else
			{
				static if (is(T : mt))
				{
					matrix[i / cols][i % cols] = head;
					construct!(i + 1)(tail);
				}
				else
				{
					static if (is(T == Vector!(mt, cols)))
					{
						static if (i % cols == 0)
						{
							matrix[i / cols] = head.vector;
							construct!(i + T.dimension)(tail);
						}
						else
						{
							static assert(false, "Can't convert Vector into the matrix. Maybe it doesn't align to the columns correctly or dimension doesn't fit");
						}
					}
					else
					{
						static if (isDynamicArray!T)
						{
							foreach (j; 0 .. cols * rows)
							{
								matrix[j / cols][j % cols] = head[j];
							}
						}
						else
						{
							static assert(false, "Matrix constructor argument must be of type " ~ mt.stringof ~ " or Vector, not " ~ T.stringof);
						}
					}
				}
			}
		}
		private void construct(int i)()
		{
			static assert(i == rows * cols, "Not enough arguments passed to constructor");
		}
		this(Args...)(Args args)
		{
			construct!0(args);
		}
		this(T)(T mat) if (is_matrix!T && T.cols >= cols && T.rows >= rows)
		{
			foreach (r; TupleRange!(0, rows))
			{
				foreach (c; TupleRange!(0, cols))
				{
					matrix[r][c] = mat.matrix[r][c];
				}
			}
		}
		this(T)(T mat) if (is_matrix!T && T.cols < cols && T.rows < rows)
		{
			make_identity();
			foreach (r; TupleRange!(0, T.rows))
			{
				foreach (c; TupleRange!(0, T.cols))
				{
					matrix[r][c] = mat.matrix[r][c];
				}
			}
		}
		this()(mt value)
		{
			clear(value);
		}
		const @property bool isFinite()
		{
			static if (isIntegral!type)
			{
				return true;
			}
			else
			{
				foreach (row; matrix)
				{
					foreach (col; row)
					{
						if (isNaN(col) || isInfinity(col))
						{
							return false;
						}
					}
				}
				return true;
			}
		}
		deprecated("Use isFinite instead of ok") alias ok = isFinite;
		void clear(mt value)
		{
			foreach (r; TupleRange!(0, rows))
			{
				foreach (c; TupleRange!(0, cols))
				{
					matrix[r][c] = value;
				}
			}
		}
		static if (rows == cols)
		{
			void make_identity()
			{
				clear(0);
				foreach (r; TupleRange!(0, rows))
				{
					matrix[r][r] = 1;
				}
			}
			static @property Matrix identity()
			{
				Matrix ret;
				ret.clear(0);
				foreach (r; TupleRange!(0, rows))
				{
					ret.matrix[r][r] = 1;
				}
				return ret;
			}
			void transpose()
			{
				matrix = transposed().matrix;
			}
		}
		const @property Matrix!(mt, cols, rows) transposed()
		{
			typeof(return) ret;
			foreach (r; TupleRange!(0, rows))
			{
				foreach (c; TupleRange!(0, cols))
				{
					ret.matrix[c][r] = matrix[r][c];
				}
			}
			return ret;
		}
		static if (rows == 2 && cols == 2)
		{
			const @property mt det()
			{
				return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0];
			}
			private const Matrix invert(ref Matrix mat)
			{
				static if (isFloatingPoint!mt && rmul)
				{
					mt d = 1 / det;
					mat.matrix = [[matrix[1][1] * d, -matrix[0][1] * d], [-matrix[1][0] * d, matrix[0][0] * d]];
				}
				else
				{
					mt d = det;
					mat.matrix = [[matrix[1][1] / d, -matrix[0][1] / d], [-matrix[1][0] / d, matrix[0][0] / d]];
				}
				return mat;
			}
			static Matrix scaling(mt x, mt y)
			{
				Matrix ret = Matrix.identity;
				ret.matrix[0][0] = x;
				ret.matrix[1][1] = y;
				return ret;
			}
			Matrix scale(mt x, mt y)
			{
				this = Matrix.scaling(x, y) * this;
				return this;
			}
		}
		else
		{
			static if (rows == 3 && cols == 3)
			{
				const @property mt det()
				{
					return matrix[0][0] * matrix[1][1] * matrix[2][2] + matrix[0][1] * matrix[1][2] * matrix[2][0] + matrix[0][2] * matrix[1][0] * matrix[2][1] - matrix[0][2] * matrix[1][1] * matrix[2][0] - matrix[0][1] * matrix[1][0] * matrix[2][2] - matrix[0][0] * matrix[1][2] * matrix[2][1];
				}
				private const Matrix invert(ref Matrix mat)
				{
					static if (isFloatingPoint!mt && rmul)
					{
						mt d = 1 / det;
						enum op = "*";
					}
					else
					{
						mt d = det;
						enum op = "/";
					}
					mixin("\x0a            mat.matrix = [[(matrix[1][1] * matrix[2][2] - matrix[1][2] * matrix[2][1])" ~ op ~ "d,\x0a                           (matrix[0][2] * matrix[2][1] - matrix[0][1] * matrix[2][2])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[1][2] - matrix[0][2] * matrix[1][1])" ~ op ~ "d],\x0a                          [(matrix[1][2] * matrix[2][0] - matrix[1][0] * matrix[2][2])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][2] - matrix[0][2] * matrix[2][0])" ~ op ~ "d,\x0a                           (matrix[0][2] * matrix[1][0] - matrix[0][0] * matrix[1][2])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][1] - matrix[1][1] * matrix[2][0])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[2][0] - matrix[0][0] * matrix[2][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0])" ~ op ~ "d]];\x0a            ");
					return mat;
				}
			}
			else
			{
				static if (rows == 4 && cols == 4)
				{
					const @property mt det()
					{
						return matrix[0][3] * matrix[1][2] * matrix[2][1] * matrix[3][0] - matrix[0][2] * matrix[1][3] * matrix[2][1] * matrix[3][0] - matrix[0][3] * matrix[1][1] * matrix[2][2] * matrix[3][0] + matrix[0][1] * matrix[1][3] * matrix[2][2] * matrix[3][0] + matrix[0][2] * matrix[1][1] * matrix[2][3] * matrix[3][0] - matrix[0][1] * matrix[1][2] * matrix[2][3] * matrix[3][0] - matrix[0][3] * matrix[1][2] * matrix[2][0] * matrix[3][1] + matrix[0][2] * matrix[1][3] * matrix[2][0] * matrix[3][1] + matrix[0][3] * matrix[1][0] * matrix[2][2] * matrix[3][1] - matrix[0][0] * matrix[1][3] * matrix[2][2] * matrix[3][1] - matrix[0][2] * matrix[1][0] * matrix[2][3] * matrix[3][1] + matrix[0][0] * matrix[1][2] * matrix[2][3] * matrix[3][1] + matrix[0][3] * matrix[1][1] * matrix[2][0] * matrix[3][2] - matrix[0][1] * matrix[1][3] * matrix[2][0] * matrix[3][2] - matrix[0][3] * matrix[1][0] * matrix[2][1] * matrix[3][2] + matrix[0][0] * matrix[1][3] * matrix[2][1] * matrix[3][2] + matrix[0][1] * matrix[1][0] * matrix[2][3] * matrix[3][2] - matrix[0][0] * matrix[1][1] * matrix[2][3] * matrix[3][2] - matrix[0][2] * matrix[1][1] * matrix[2][0] * matrix[3][3] + matrix[0][1] * matrix[1][2] * matrix[2][0] * matrix[3][3] + matrix[0][2] * matrix[1][0] * matrix[2][1] * matrix[3][3] - matrix[0][0] * matrix[1][2] * matrix[2][1] * matrix[3][3] - matrix[0][1] * matrix[1][0] * matrix[2][2] * matrix[3][3] + matrix[0][0] * matrix[1][1] * matrix[2][2] * matrix[3][3];
					}
					private const Matrix invert(ref Matrix mat)
					{
						static if (isFloatingPoint!mt && rmul)
						{
							mt d = 1 / det;
							enum op = "*";
						}
						else
						{
							mt d = det;
							enum op = "/";
						}
						mixin("\x0a            mat.matrix = [[(matrix[1][1] * matrix[2][2] * matrix[3][3] + matrix[1][2] * matrix[2][3] * matrix[3][1] + matrix[1][3] * matrix[2][1] * matrix[3][2]\x0a                          - matrix[1][1] * matrix[2][3] * matrix[3][2] - matrix[1][2] * matrix[2][1] * matrix[3][3] - matrix[1][3] * matrix[2][2] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[2][3] * matrix[3][2] + matrix[0][2] * matrix[2][1] * matrix[3][3] + matrix[0][3] * matrix[2][2] * matrix[3][1]\x0a                          - matrix[0][1] * matrix[2][2] * matrix[3][3] - matrix[0][2] * matrix[2][3] * matrix[3][1] - matrix[0][3] * matrix[2][1] * matrix[3][2])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[1][2] * matrix[3][3] + matrix[0][2] * matrix[1][3] * matrix[3][1] + matrix[0][3] * matrix[1][1] * matrix[3][2]\x0a                          - matrix[0][1] * matrix[1][3] * matrix[3][2] - matrix[0][2] * matrix[1][1] * matrix[3][3] - matrix[0][3] * matrix[1][2] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[1][3] * matrix[2][2] + matrix[0][2] * matrix[1][1] * matrix[2][3] + matrix[0][3] * matrix[1][2] * matrix[2][1]\x0a                          - matrix[0][1] * matrix[1][2] * matrix[2][3] - matrix[0][2] * matrix[1][3] * matrix[2][1] - matrix[0][3] * matrix[1][1] * matrix[2][2])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][3] * matrix[3][2] + matrix[1][2] * matrix[2][0] * matrix[3][3] + matrix[1][3] * matrix[2][2] * matrix[3][0]\x0a                          - matrix[1][0] * matrix[2][2] * matrix[3][3] - matrix[1][2] * matrix[2][3] * matrix[3][0] - matrix[1][3] * matrix[2][0] * matrix[3][2])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][2] * matrix[3][3] + matrix[0][2] * matrix[2][3] * matrix[3][0] + matrix[0][3] * matrix[2][0] * matrix[3][2]\x0a                          - matrix[0][0] * matrix[2][3] * matrix[3][2] - matrix[0][2] * matrix[2][0] * matrix[3][3] - matrix[0][3] * matrix[2][2] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][3] * matrix[3][2] + matrix[0][2] * matrix[1][0] * matrix[3][3] + matrix[0][3] * matrix[1][2] * matrix[3][0]\x0a                          - matrix[0][0] * matrix[1][2] * matrix[3][3] - matrix[0][2] * matrix[1][3] * matrix[3][0] - matrix[0][3] * matrix[1][0] * matrix[3][2])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][2] * matrix[2][3] + matrix[0][2] * matrix[1][3] * matrix[2][0] + matrix[0][3] * matrix[1][0] * matrix[2][2]\x0a                          - matrix[0][0] * matrix[1][3] * matrix[2][2] - matrix[0][2] * matrix[1][0] * matrix[2][3] - matrix[0][3] * matrix[1][2] * matrix[2][0])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][1] * matrix[3][3] + matrix[1][1] * matrix[2][3] * matrix[3][0] + matrix[1][3] * matrix[2][0] * matrix[3][1]\x0a                          - matrix[1][0] * matrix[2][3] * matrix[3][1] - matrix[1][1] * matrix[2][0] * matrix[3][3] - matrix[1][3] * matrix[2][1] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][3] * matrix[3][1] + matrix[0][1] * matrix[2][0] * matrix[3][3] + matrix[0][3] * matrix[2][1] * matrix[3][0]\x0a                          - matrix[0][0] * matrix[2][1] * matrix[3][3] - matrix[0][1] * matrix[2][3] * matrix[3][0] - matrix[0][3] * matrix[2][0] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][1] * matrix[3][3] + matrix[0][1] * matrix[1][3] * matrix[3][0] + matrix[0][3] * matrix[1][0] * matrix[3][1]\x0a                          - matrix[0][0] * matrix[1][3] * matrix[3][1] - matrix[0][1] * matrix[1][0] * matrix[3][3] - matrix[0][3] * matrix[1][1] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][3] * matrix[2][1] + matrix[0][1] * matrix[1][0] * matrix[2][3] + matrix[0][3] * matrix[1][1] * matrix[2][0]\x0a                          - matrix[0][0] * matrix[1][1] * matrix[2][3] - matrix[0][1] * matrix[1][3] * matrix[2][0] - matrix[0][3] * matrix[1][0] * matrix[2][1])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][2] * matrix[3][1] + matrix[1][1] * matrix[2][0] * matrix[3][2] + matrix[1][2] * matrix[2][1] * matrix[3][0]\x0a                          - matrix[1][0] * matrix[2][1] * matrix[3][2] - matrix[1][1] * matrix[2][2] * matrix[3][0] - matrix[1][2] * matrix[2][0] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][1] * matrix[3][2] + matrix[0][1] * matrix[2][2] * matrix[3][0] + matrix[0][2] * matrix[2][0] * matrix[3][1]\x0a                          - matrix[0][0] * matrix[2][2] * matrix[3][1] - matrix[0][1] * matrix[2][0] * matrix[3][2] - matrix[0][2] * matrix[2][1] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][2] * matrix[3][1] + matrix[0][1] * matrix[1][0] * matrix[3][2] + matrix[0][2] * matrix[1][1] * matrix[3][0]\x0a                          - matrix[0][0] * matrix[1][1] * matrix[3][2] - matrix[0][1] * matrix[1][2] * matrix[3][0] - matrix[0][2] * matrix[1][0] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][1] * matrix[2][2] + matrix[0][1] * matrix[1][2] * matrix[2][0] + matrix[0][2] * matrix[1][0] * matrix[2][1]\x0a                          - matrix[0][0] * matrix[1][2] * matrix[2][1] - matrix[0][1] * matrix[1][0] * matrix[2][2] - matrix[0][2] * matrix[1][1] * matrix[2][0])" ~ op ~ "d]];\x0a            ");
						return mat;
					}
					static if (isFloatingPoint!mt)
					{
						private static mt[6] cperspective(mt width, mt height, mt fov, mt near, mt far)
						in
						{
							assert(height != 0);
						}
						body
						{
							mt aspect = width / height;
							mt top = near * tan(fov * (PI / 360.000));
							mt bottom = -top;
							mt right = top * aspect;
							mt left = -right;
							return [left, right, bottom, top, near, far];
						}
						static Matrix perspective(mt width, mt height, mt fov, mt near, mt far)
						{
							mt[6] cdata = cperspective(width, height, fov, near, far);
							return perspective(cdata[0], cdata[1], cdata[2], cdata[3], cdata[4], cdata[5]);
						}
						static Matrix perspective(mt left, mt right, mt bottom, mt top, mt near, mt far)
						in
						{
							assert(right - left != 0);
							assert(top - bottom != 0);
							assert(far - near != 0);
						}
						body
						{
							Matrix ret;
							ret.clear(0);
							ret.matrix[0][0] = 2 * near / (right - left);
							ret.matrix[0][2] = (right + left) / (right - left);
							ret.matrix[1][1] = 2 * near / (top - bottom);
							ret.matrix[1][2] = (top + bottom) / (top - bottom);
							ret.matrix[2][2] = -(far + near) / (far - near);
							ret.matrix[2][3] = -(2 * far * near) / (far - near);
							ret.matrix[3][2] = -1;
							return ret;
						}
						static Matrix perspective_inverse(mt width, mt height, mt fov, mt near, mt far)
						{
							mt[6] cdata = cperspective(width, height, fov, near, far);
							return perspective_inverse(cdata[0], cdata[1], cdata[2], cdata[3], cdata[4], cdata[5]);
						}
						static Matrix perspective_inverse(mt left, mt right, mt bottom, mt top, mt near, mt far)
						in
						{
							assert(near != 0);
							assert(far != 0);
						}
						body
						{
							Matrix ret;
							ret.clear(0);
							ret.matrix[0][0] = (right - left) / (2 * near);
							ret.matrix[0][3] = (right + left) / (2 * near);
							ret.matrix[1][1] = (top - bottom) / (2 * near);
							ret.matrix[1][3] = (top + bottom) / (2 * near);
							ret.matrix[2][3] = -1;
							ret.matrix[3][2] = -(far - near) / (2 * far * near);
							ret.matrix[3][3] = (far + near) / (2 * far * near);
							return ret;
						}
						static Matrix orthographic(mt left, mt right, mt bottom, mt top, mt near, mt far)
						in
						{
							assert(right - left != 0);
							assert(top - bottom != 0);
							assert(far - near != 0);
						}
						body
						{
							Matrix ret;
							ret.clear(0);
							ret.matrix[0][0] = 2 / (right - left);
							ret.matrix[0][3] = -(right + left) / (right - left);
							ret.matrix[1][1] = 2 / (top - bottom);
							ret.matrix[1][3] = -(top + bottom) / (top - bottom);
							ret.matrix[2][2] = -2 / (far - near);
							ret.matrix[2][3] = -(far + near) / (far - near);
							ret.matrix[3][3] = 1;
							return ret;
						}
						static Matrix orthographic_inverse(mt left, mt right, mt bottom, mt top, mt near, mt far)
						{
							Matrix ret;
							ret.clear(0);
							ret.matrix[0][0] = (right - left) / 2;
							ret.matrix[0][3] = (right + left) / 2;
							ret.matrix[1][1] = (top - bottom) / 2;
							ret.matrix[1][3] = (top + bottom) / 2;
							ret.matrix[2][2] = (far - near) / -2;
							ret.matrix[2][3] = (far + near) / 2;
							ret.matrix[3][3] = 1;
							return ret;
						}
						static Matrix look_at(Vector!(mt, 3) eye, Vector!(mt, 3) target, Vector!(mt, 3) up)
						{
							alias vec3mt = Vector!(mt, 3);
							vec3mt look_dir = (target - eye).normalized;
							vec3mt up_dir = up.normalized;
							vec3mt right_dir = cross(look_dir, up_dir).normalized;
							vec3mt perp_up_dir = cross(right_dir, look_dir);
							Matrix ret = Matrix.identity;
							ret.matrix[0][0..3] = right_dir.vector[];
							ret.matrix[1][0..3] = perp_up_dir.vector[];
							ret.matrix[2][0..3] = (-look_dir).vector[];
							ret.matrix[0][3] = -dot(eye, right_dir);
							ret.matrix[1][3] = -dot(eye, perp_up_dir);
							ret.matrix[2][3] = dot(eye, look_dir);
							return ret;
						}
					}
				}
			}
		}
		static if (rows == cols && rows >= 3 && rows <= 4)
		{
			static Matrix translation(mt x, mt y, mt z)
			{
				Matrix ret = Matrix.identity;
				ret.matrix[0][cols - 1] = x;
				ret.matrix[1][cols - 1] = y;
				ret.matrix[2][cols - 1] = z;
				return ret;
			}
			static Matrix translation(Vector!(mt, 3) v)
			{
				Matrix ret = Matrix.identity;
				ret.matrix[0][cols - 1] = v.x;
				ret.matrix[1][cols - 1] = v.y;
				ret.matrix[2][cols - 1] = v.z;
				return ret;
			}
			Matrix translate(mt x, mt y, mt z)
			{
				this = Matrix.translation(x, y, z) * this;
				return this;
			}
			Matrix translate(Vector!(mt, 3) v)
			{
				this = Matrix.translation(v) * this;
				return this;
			}
			static Matrix scaling(mt x, mt y, mt z)
			{
				Matrix ret = Matrix.identity;
				ret.matrix[0][0] = x;
				ret.matrix[1][1] = y;
				ret.matrix[2][2] = z;
				return ret;
			}
			Matrix scale(mt x, mt y, mt z)
			{
				this = Matrix.scaling(x, y, z) * this;
				return this;
			}
		}
		static if (rows == cols && rows >= 3)
		{
			static if (isFloatingPoint!mt)
			{
				static Matrix rotation(real alpha, Vector!(mt, 3) axis)
				{
					Matrix mult = Matrix.identity;
					if (axis.length != 1)
					{
						axis.normalize();
					}
					real cosa = cos(alpha);
					real sina = sin(alpha);
					Vector!(mt, 3) temp = (1 - cosa) * axis;
					mult.matrix[0][0] = to!mt(cosa + temp.x * axis.x);
					mult.matrix[0][1] = to!mt(temp.x * axis.y + sina * axis.z);
					mult.matrix[0][2] = to!mt(temp.x * axis.z - sina * axis.y);
					mult.matrix[1][0] = to!mt(temp.y * axis.x - sina * axis.z);
					mult.matrix[1][1] = to!mt(cosa + temp.y * axis.y);
					mult.matrix[1][2] = to!mt(temp.y * axis.z + sina * axis.x);
					mult.matrix[2][0] = to!mt(temp.z * axis.x + sina * axis.y);
					mult.matrix[2][1] = to!mt(temp.z * axis.y - sina * axis.x);
					mult.matrix[2][2] = to!mt(cosa + temp.z * axis.z);
					return mult;
				}
				static Matrix rotation(real alpha, mt x, mt y, mt z)
				{
					return Matrix.rotation(alpha, Vector!(mt, 3)(x, y, z));
				}
				static Matrix xrotation(real alpha)
				{
					Matrix mult = Matrix.identity;
					mt cosamt = to!mt(cos(alpha));
					mt sinamt = to!mt(sin(alpha));
					mult.matrix[1][1] = cosamt;
					mult.matrix[1][2] = -sinamt;
					mult.matrix[2][1] = sinamt;
					mult.matrix[2][2] = cosamt;
					return mult;
				}
				static Matrix yrotation(real alpha)
				{
					Matrix mult = Matrix.identity;
					mt cosamt = to!mt(cos(alpha));
					mt sinamt = to!mt(sin(alpha));
					mult.matrix[0][0] = cosamt;
					mult.matrix[0][2] = sinamt;
					mult.matrix[2][0] = -sinamt;
					mult.matrix[2][2] = cosamt;
					return mult;
				}
				static Matrix zrotation(real alpha)
				{
					Matrix mult = Matrix.identity;
					mt cosamt = to!mt(cos(alpha));
					mt sinamt = to!mt(sin(alpha));
					mult.matrix[0][0] = cosamt;
					mult.matrix[0][1] = -sinamt;
					mult.matrix[1][0] = sinamt;
					mult.matrix[1][1] = cosamt;
					return mult;
				}
				Matrix rotate(real alpha, Vector!(mt, 3) axis)
				{
					this = rotation(alpha, axis) * this;
					return this;
				}
				Matrix rotatex(real alpha)
				{
					this = xrotation(alpha) * this;
					return this;
				}
				Matrix rotatey(real alpha)
				{
					this = yrotation(alpha) * this;
					return this;
				}
				Matrix rotatez(real alpha)
				{
					this = zrotation(alpha) * this;
					return this;
				}
			}
			void set_translation(mt[] values...)
			in
			{
				assert(values.length >= rows - 1);
			}
			body
			{
				foreach (r; TupleRange!(0, rows - 1))
				{
					matrix[r][rows - 1] = values[r];
				}
			}
			void set_translation(Matrix mat)
			{
				foreach (r; TupleRange!(0, rows - 1))
				{
					matrix[r][rows - 1] = mat.matrix[r][rows - 1];
				}
			}
			Matrix get_translation()
			{
				Matrix ret = Matrix.identity;
				foreach (r; TupleRange!(0, rows - 1))
				{
					ret.matrix[r][rows - 1] = matrix[r][rows - 1];
				}
				return ret;
			}
			void set_scale(mt[] values...)
			in
			{
				assert(values.length >= rows - 1);
			}
			body
			{
				foreach (r; TupleRange!(0, rows - 1))
				{
					matrix[r][r] = values[r];
				}
			}
			void set_scale(Matrix mat)
			{
				foreach (r; TupleRange!(0, rows - 1))
				{
					matrix[r][r] = mat.matrix[r][r];
				}
			}
			Matrix get_scale()
			{
				Matrix ret = Matrix.identity;
				foreach (r; TupleRange!(0, rows - 1))
				{
					ret.matrix[r][r] = matrix[r][r];
				}
				return ret;
			}
			void set_rotation(Matrix!(mt, 3, 3) rot)
			{
				foreach (r; TupleRange!(0, 3))
				{
					foreach (c; TupleRange!(0, 3))
					{
						matrix[r][c] = rot[r][c];
					}
				}
			}
			Matrix!(mt, 3, 3) get_rotation()
			{
				Matrix!(mt, 3, 3) ret = Matrix!(mt, 3, 3).identity;
				foreach (r; TupleRange!(0, 3))
				{
					foreach (c; TupleRange!(0, 3))
					{
						ret.matrix[r][c] = matrix[r][c];
					}
				}
				return ret;
			}
		}
		static if (rows == cols && rows >= 2 && rows <= 4)
		{
			const @property Matrix inverse()
			{
				Matrix mat;
				invert(mat);
				return mat;
			}
			void invert()
			{
				Matrix temp;
				invert(temp);
				this.matrix = temp.matrix;
			}
		}
		private const void mms(mt inp, ref Matrix mat)
		{
			for (int r = 0;
			 r < rows; r++)
			{
				{
					for (int c = 0;
					 c < cols; c++)
					{
						{
							mat.matrix[r][c] = matrix[r][c] * inp;
						}
					}
				}
			}
		}
		private const void masm(string op)(Matrix inp, ref Matrix mat)
		{
			foreach (r; TupleRange!(0, rows))
			{
				foreach (c; TupleRange!(0, cols))
				{
					mat.matrix[r][c] = mixin("inp.matrix[r][c]" ~ op ~ "matrix[r][c]");
				}
			}
		}
		const Matrix!(mt, rows, T.cols) opBinary(string op : "*", T)(T inp) if (isCompatibleMatrix!T && T.rows == cols)
		{
			Matrix!(mt, rows, T.cols) ret;
			foreach (r; TupleRange!(0, rows))
			{
				foreach (c; TupleRange!(0, T.cols))
				{
					ret.matrix[r][c] = 0;
					foreach (c2; TupleRange!(0, cols))
					{
						ret.matrix[r][c] += matrix[r][c2] * inp.matrix[c2][c];
					}
				}
			}
			return ret;
		}
		const Vector!(mt, rows) opBinary(string op : "*", T : Vector!(mt, cols))(T inp)
		{
			Vector!(mt, rows) ret;
			ret.clear(0);
			foreach (c; TupleRange!(0, cols))
			{
				foreach (r; TupleRange!(0, rows))
				{
					ret.vector[r] += matrix[r][c] * inp.vector[c];
				}
			}
			return ret;
		}
		const Matrix opBinary(string op : "*")(mt inp)
		{
			Matrix ret;
			mms(inp, ret);
			return ret;
		}
		const Matrix opBinaryRight(string op : "*")(mt inp)
		{
			return this.opBinary!op(inp);
		}
		const Matrix opBinary(string op)(Matrix inp) if (op == "+" || op == "-")
		{
			Matrix ret;
			masm!op(inp, ret);
			return ret;
		}
		void opOpAssign(string op : "*")(mt inp)
		{
			mms(inp, this);
		}
		void opOpAssign(string op)(Matrix inp) if (op == "+" || op == "-")
		{
			masm!op(inp, this);
		}
		void opOpAssign(string op)(Matrix inp) if (op == "*")
		{
			this = this * inp;
		}
		const bool opCast(T : bool)()
		{
			return isFinite;
		}
	}
}
alias mat2 = Matrix!(float, 2, 2);
alias mat3 = Matrix!(float, 3, 3);
alias mat34 = Matrix!(float, 3, 4);
alias mat4 = Matrix!(float, 4, 4);
private {}
struct Quaternion(type)
{
	alias qt = type;
	qt[4] quaternion;
	auto const @property value_ptr()
	{
		return quaternion.ptr;
	}
	@property string as_string()
	{
		return format("%s", quaternion);
	}
	alias toString = as_string;
	nothrow pure @safe 
	{
		private const @property qt get_(char coord)()
		{
			return quaternion[coord_to_index!coord];
		}
		private @property void set_(char coord)(qt value)
		{
			quaternion[coord_to_index!coord] = value;
		}
		alias w = get_!'w';
		alias w = set_!'w';
		alias x = get_!'x';
		alias x = set_!'x';
		alias y = get_!'y';
		alias y = set_!'y';
		alias z = get_!'z';
		alias z = set_!'z';
		this(qt w_, qt x_, qt y_, qt z_)
		{
			w = w_;
			x = x_;
			y = y_;
			z = z_;
		}
		this(qt w_, Vector!(qt, 3) vec)
		{
			w = w_;
			quaternion[1..4] = vec.vector[];
		}
		this(Vector!(qt, 4) vec)
		{
			quaternion[] = vec.vector[];
		}
		const @property bool isFinite()
		{
			foreach (q; quaternion)
			{
				if (isNaN(q) || isInfinity(q))
				{
					return false;
				}
			}
			return true;
		}
		deprecated("Use isFinite instead of ok") alias ok = isFinite;
		template coord_to_index(char c)
		{
			static if (c == 'w')
			{
				enum coord_to_index = 0;
			}
			else
			{
				static if (c == 'x')
				{
					enum coord_to_index = 1;
				}
				else
				{
					static if (c == 'y')
					{
						enum coord_to_index = 2;
					}
					else
					{
						static if (c == 'z')
						{
							enum coord_to_index = 3;
						}
						else
						{
							static assert(false, "accepted coordinates are x, y, z and w not " ~ c ~ ".");
						}
					}
				}
			}
		}
		const @property real magnitude_squared()
		{
			return to!real(w ^^ 2 + x ^^ 2 + y ^^ 2 + z ^^ 2);
		}
		const @property real magnitude()
		{
			return sqrt(magnitude_squared);
		}
		static @property Quaternion identity()
		{
			return Quaternion(1, 0, 0, 0);
		}
		void make_identity()
		{
			w = 1;
			x = 0;
			y = 0;
			z = 0;
		}
		void invert()
		{
			x = -x;
			y = -y;
			z = -z;
		}
		alias conjugate = invert;
		const @property Quaternion inverse()
		{
			return Quaternion(w, -x, -y, -z);
		}
		alias conjugated = inverse;
		static Quaternion from_matrix(Matrix!(qt, 3, 3) matrix)
		{
			Quaternion ret;
			auto mat = matrix.matrix;
			qt trace = mat[0][0] + mat[1][1] + mat[2][2];
			if (trace > 0)
			{
				real s = 0.5 / sqrt(trace + 1.00000F);
				ret.w = to!qt(0.25 / s);
				ret.x = to!qt((mat[2][1] - mat[1][2]) * s);
				ret.y = to!qt((mat[0][2] - mat[2][0]) * s);
				ret.z = to!qt((mat[1][0] - mat[0][1]) * s);
			}
			else
				if (mat[0][0] > mat[1][1] && mat[0][0] > mat[2][2])
				{
					real s = 2.00000 * sqrt(1.00000 + mat[0][0] - mat[1][1] - mat[2][2]);
					ret.w = to!qt((mat[2][1] - mat[1][2]) / s);
					ret.x = to!qt(0.25F * s);
					ret.y = to!qt((mat[0][1] + mat[1][0]) / s);
					ret.z = to!qt((mat[0][2] + mat[2][0]) / s);
				}
				else
					if (mat[1][1] > mat[2][2])
					{
						real s = 2.00000 * sqrt(1 + mat[1][1] - mat[0][0] - mat[2][2]);
						ret.w = to!qt((mat[0][2] - mat[2][0]) / s);
						ret.x = to!qt((mat[0][1] + mat[1][0]) / s);
						ret.y = to!qt(0.25F * s);
						ret.z = to!qt((mat[1][2] + mat[2][1]) / s);
					}
					else
					{
						real s = 2.00000 * sqrt(1 + mat[2][2] - mat[0][0] - mat[1][1]);
						ret.w = to!qt((mat[1][0] - mat[0][1]) / s);
						ret.x = to!qt((mat[0][2] + mat[2][0]) / s);
						ret.y = to!qt((mat[1][2] + mat[2][1]) / s);
						ret.z = to!qt(0.25F * s);
					}
			return ret;
		}
		const Matrix!(qt, rows, cols) to_matrix(int rows, int cols)() if (rows >= 3 && cols >= 3)
		{
			static if (rows == 3 && cols == 3)
			{
				Matrix!(qt, rows, cols) ret;
			}
			else
			{
				Matrix!(qt, rows, cols) ret = Matrix!(qt, rows, cols).identity;
			}
			qt xx = x ^^ 2;
			qt xy = x * y;
			qt xz = x * z;
			qt xw = x * w;
			qt yy = y ^^ 2;
			qt yz = y * z;
			qt yw = y * w;
			qt zz = z ^^ 2;
			qt zw = z * w;
			ret.matrix[0][0] = 1 - 2 * (yy + zz);
			ret.matrix[0][1] = 2 * (xy - zw);
			ret.matrix[0][2] = 2 * (xz + yw);
			ret.matrix[1][0] = 2 * (xy + zw);
			ret.matrix[1][1] = 1 - 2 * (xx + zz);
			ret.matrix[1][2] = 2 * (yz - xw);
			ret.matrix[2][0] = 2 * (xz - yw);
			ret.matrix[2][1] = 2 * (yz + xw);
			ret.matrix[2][2] = 1 - 2 * (xx + yy);
			return ret;
		}
		void normalize()
		{
			qt m = to!qt(magnitude);
			if (m != 0)
			{
				w = w / m;
				x = x / m;
				y = y / m;
				z = z / m;
			}
		}
		const Quaternion normalized()
		{
			Quaternion ret;
			qt m = to!qt(magnitude);
			if (m != 0)
			{
				ret.w = w / m;
				ret.x = x / m;
				ret.y = y / m;
				ret.z = z / m;
			}
			else
			{
				ret = Quaternion(w, x, y, z);
			}
			return ret;
		}
		const @property real yaw()
		{
			return atan2(to!real(2.00000 * (w * z + x * y)), to!real(1.00000 - 2.00000 * (y * y + z * z)));
		}
		const @property real pitch()
		{
			return asin(to!real(2.00000 * (w * y - z * x)));
		}
		const @property real roll()
		{
			return atan2(to!real(2.00000 * (w * x + y * z)), to!real(1.00000 - 2.00000 * (x * x + y * y)));
		}
		static Quaternion xrotation(real alpha)
		{
			Quaternion ret;
			alpha /= 2;
			ret.w = to!qt(cos(alpha));
			ret.x = to!qt(sin(alpha));
			ret.y = 0;
			ret.z = 0;
			return ret;
		}
		static Quaternion yrotation(real alpha)
		{
			Quaternion ret;
			alpha /= 2;
			ret.w = to!qt(cos(alpha));
			ret.x = 0;
			ret.y = to!qt(sin(alpha));
			ret.z = 0;
			return ret;
		}
		static Quaternion zrotation(real alpha)
		{
			Quaternion ret;
			alpha /= 2;
			ret.w = to!qt(cos(alpha));
			ret.x = 0;
			ret.y = 0;
			ret.z = to!qt(sin(alpha));
			return ret;
		}
		static Quaternion axis_rotation(real alpha, Vector!(qt, 3) axis)
		{
			if (alpha == 0)
			{
				return Quaternion.identity;
			}
			Quaternion ret;
			alpha /= 2;
			qt sinaqt = to!qt(sin(alpha));
			ret.w = to!qt(cos(alpha));
			ret.x = axis.x * sinaqt;
			ret.y = axis.y * sinaqt;
			ret.z = axis.z * sinaqt;
			return ret;
		}
		static Quaternion euler_rotation(real roll, real pitch, real yaw)
		{
			Quaternion ret;
			auto cr = cos(roll / 2.00000);
			auto cp = cos(pitch / 2.00000);
			auto cy = cos(yaw / 2.00000);
			auto sr = sin(roll / 2.00000);
			auto sp = sin(pitch / 2.00000);
			auto sy = sin(yaw / 2.00000);
			ret.w = cr * cp * cy + sr * sp * sy;
			ret.x = sr * cp * cy - cr * sp * sy;
			ret.y = cr * sp * cy + sr * cp * sy;
			ret.z = cr * cp * sy - sr * sp * cy;
			return ret;
		}
		Quaternion rotatex(real alpha)
		{
			this = xrotation(alpha) * this;
			return this;
		}
		Quaternion rotatey(real alpha)
		{
			this = yrotation(alpha) * this;
			return this;
		}
		Quaternion rotatez(real alpha)
		{
			this = zrotation(alpha) * this;
			return this;
		}
		Quaternion rotate_axis(real alpha, Vector!(qt, 3) axis)
		{
			this = axis_rotation(alpha, axis) * this;
			return this;
		}
		Quaternion rotate_euler(real heading, real attitude, real bank)
		{
			this = euler_rotation(heading, attitude, bank) * this;
			return this;
		}
		const Quaternion opBinary(string op : "*")(Quaternion inp)
		{
			Quaternion ret;
			ret.w = -x * inp.x - y * inp.y - z * inp.z + w * inp.w;
			ret.x = x * inp.w + y * inp.z - z * inp.y + w * inp.x;
			ret.y = -x * inp.z + y * inp.w + z * inp.x + w * inp.y;
			ret.z = x * inp.y - y * inp.x + z * inp.w + w * inp.z;
			return ret;
		}
		auto const opBinaryRight(string op, T)(T inp) if (!is_quaternion!T)
		{
			return this.opBinary!op(inp);
		}
		const Quaternion opBinary(string op)(Quaternion inp) if (op == "+" || op == "-")
		{
			Quaternion ret;
			mixin("ret.w = w" ~ op ~ "inp.w;");
			mixin("ret.x = x" ~ op ~ "inp.x;");
			mixin("ret.y = y" ~ op ~ "inp.y;");
			mixin("ret.z = z" ~ op ~ "inp.z;");
			return ret;
		}
		const Vector!(qt, 3) opBinary(string op : "*")(Vector!(qt, 3) inp)
		{
			Vector!(qt, 3) ret;
			qt ww = w ^^ 2;
			qt w2 = w * 2;
			qt wx2 = w2 * x;
			qt wy2 = w2 * y;
			qt wz2 = w2 * z;
			qt xx = x ^^ 2;
			qt x2 = x * 2;
			qt xy2 = x2 * y;
			qt xz2 = x2 * z;
			qt yy = y ^^ 2;
			qt yz2 = 2 * y * z;
			qt zz = z * z;
			ret.vector = [ww * inp.x + wy2 * inp.z - wz2 * inp.y + xx * inp.x + xy2 * inp.y + xz2 * inp.z - zz * inp.x - yy * inp.x, xy2 * inp.x + yy * inp.y + yz2 * inp.z + wz2 * inp.x - zz * inp.y + ww * inp.y - wx2 * inp.z - xx * inp.y, xz2 * inp.x + yz2 * inp.y + zz * inp.z - wy2 * inp.x - yy * inp.z + wx2 * inp.y - xx * inp.z + ww * inp.z];
			return ret;
		}
		const Quaternion opBinary(string op : "*")(qt inp)
		{
			return Quaternion(w * inp, x * inp, y * inp, z * inp);
		}
		void opOpAssign(string op : "*")(Quaternion inp)
		{
			qt w2 = -x * inp.x - y * inp.y - z * inp.z + w * inp.w;
			qt x2 = x * inp.w + y * inp.z - z * inp.y + w * inp.x;
			qt y2 = -x * inp.z + y * inp.w + z * inp.x + w * inp.y;
			qt z2 = x * inp.y - y * inp.x + z * inp.w + w * inp.z;
			w = w2;
			x = x2;
			y = y2;
			z = z2;
		}
		void opOpAssign(string op)(Quaternion inp) if (op == "+" || op == "-")
		{
			mixin("w = w" ~ op ~ "inp.w;");
			mixin("x = x" ~ op ~ "inp.x;");
			mixin("y = y" ~ op ~ "inp.y;");
			mixin("z = z" ~ op ~ "inp.z;");
		}
		void opOpAssign(string op : "*")(qt inp)
		{
			quaternion[0] *= inp;
			quaternion[1] *= inp;
			quaternion[2] *= inp;
			quaternion[3] *= inp;
		}
		const int opCmp(ref const Quaternion qua)
		{
			foreach (i, a; quaternion)
			{
				if (a < qua.quaternion[i])
				{
					return -1;
				}
				else
					if (a > qua.quaternion[i])
					{
						return 1;
					}
			}
			return 0;
		}
		const bool opEquals(const Quaternion qu)
		{
			return quaternion == qu.quaternion;
		}
		const bool opCast(T : bool)()
		{
			return isFinite;
		}
	}
}
alias quat = Quaternion!float;
